############################
#### Simulation example ####
############################

###################
#### load codes ###
###################
source("plfam.R")
source("simulation_settings.R")

###################################
#### generation of the example ####
###################################
n <- 200 # training sample size
nn <- 800 # testing sample size
N <- 100  # number of time points
J <- 20   # number of eigenfunctions
v <- sqrt(0.2)  # sd of error (functional predictor)
sigma <- 1      # sd of error (response)
sim.example <-sim.biexample4 # choice of example

dat <- sim.example(n=n, N=N, J=J, v=v, sigma=sigma)
tt <- dat$tt; X1 <- dat$X1; X2 <- dat$X2; U <- dat$U; y <- dat$y

##################
#### fit FPCA ####
##################
# bspline 50 knots; one smoothing parameter
fdd1 <- dat2fd2(argvals=t(tt), y=t(X1), rangeval=c(0,10), bs=NULL,
                 loglam.lims=c(-15, 5), lam.len=50, dfscale=1, nk=50)
fdd2 <- dat2fd2(argvals=t(tt), y=t(X2), rangeval=c(0,10), bs=NULL,
                 loglam.lims=c(-15, 5), lam.len=50, dfscale=1, nk=50)

fdout1 <- fdd1$fd
fdout2 <- fdd2$fd
fdoutc <- combine.fd(fdd1$fd, fdd2$fd)
lams <- c(fdd1$lambda, fdd2$lambda)

##################
#### fit zeta ####
##################
# note that some column are flipped: the sign of eigenfunction is flipped, thus
# the corresponding zetaj = 1 -zeta0j

zetalist <- list()

## separate FPCA
temp1 <- get.zeta(fdout1)
temp2 <- get.zeta(fdout2)
zetalist[[1]] <- cbind(temp1$zeta, temp2$zeta)
indd <- c(rep(1, ncol(temp1$zeta)), rep(2, ncol(temp2$zeta)))

## joint FPCA
temp <- get.zeta(fdoutc)
zetalist[[2]] <- temp$zeta

########################################
#### fit COSSO with parametric part ####
########################################
fit1list <- list()
fit2list <- list()
for (k in (1:length(zetalist))){
  fit1list[[k]] <- plcosso(y, NULL, zetalist[[k]]) # without parametric predictor
  fit2list[[k]] <- plcosso(y, U, zetalist[[k]]) # with parametric predictor
}

#################################################################
#### generation of extra data for computing prediction error ####
#################################################################
dat2 <- sim.example(n=nn, N=N, J=J, v=v, sigma=sigma)
tt2 <- dat2$tt; X21 <- dat2$X1; X22 <- dat2$X2; U2 <- dat2$U; y2 <- dat2$y


##################################
#### compute prediction error ####
##################################
zeta2list <- list()

fP1 <- fdPar(fdout1$basis, 2, lams[1])
fP2 <- fdPar(fdout2$basis, 2, lams[2])
dfd1 <- smooth.basis(argvals=t(tt2), y=t(X21), fdParobj=fP1)$fd
dfd2 <- smooth.basis(argvals=t(tt2), y=t(X22), fdParobj=fP2)$fd
dfdc <- combine.fd(dfd1, dfd2)

## separate FPCA
temp1 <- get.zeta2(dfd1, fdout1)
temp2 <- get.zeta2(dfd2, fdout2)
zeta2list[[1]] <- cbind(temp1$zeta, temp2$zeta)

## joint FPCA
temp <- get.zeta2(dfdc, fdoutc)
zeta2list[[2]] <- temp$zeta

pe1 <- array(dim=2)
pe2 <- array(dim=2)
for (k in (1:2)){
  pred1 <- predict.plcosso(fit1list[[k]], NULL, zeta2list[[k]])
  pred2 <- predict.plcosso(fit2list[[k]], U2, zeta2list[[k]])
  pe1[k] <- mean((y2-pred1)^2)
  pe2[k] <- mean((y2-pred2)^2)
}

####################
#### evaluation ####
####################

# prediction errors
pe1[1] # separate FPCA, without parametric predictor
pe1[2] # joint FPCA, without parametric predictor
pe2[1] # separate FPCA, with parametric predictor
pe2[2] # joint FPCA, with parametric predictor

# indicator of variable selection
(fit1list[[1]]$bphi>0) # separate FPCA, without parametric predictor
(fit1list[[2]]$bphi>0) # joint FPCA, without parametric predictor
(fit2list[[1]]$bphi>0) # separate FPCA, with parametric predictor
(fit2list[[2]]$bphi>0) # joint FPCA, with parametric predictor

# example of extracting regression coefficients w.r.t. the parametric predictors (including the intercept)
# # joint FPCA, with parametric predictor (fit2list[[2]])
fit2list[[2]]$btheta


# example of generating the fitted component function
# # joint FPCA, with parametric predictor (fit2list[[2]])
fs <- comps(fit2list[[2]], ngrid=1000) # evaluated at a grid of length 1000
fs$z # grid points
fs$comfit # component functions evaluated at the grid points


